local addonName, addon = ...

local ldb = LibStub:GetLibrary("LibDataBroker-1.1")

--[[-------------------------------------------------------------------------
-- Interface Options using tekConfig with Ampere's code example
-------------------------------------------------------------------------]]--

local frame = CreateFrame("Frame", nil, InterfaceOptionsFramePanelContainer)
addon.configFrame = frame
frame.name = "NinjaPanel"
frame:Hide()
frame:SetScript("OnShow", function(frame)
	-- Predeclare
	local Refresh
	local function MakeButton(parent)
		local button = CreateFrame("Button", nil, parent or frame)
		button:SetWidth(80)
		button:SetHeight(22)

		button:SetHighlightFontObject(GameFontHighlightSmall)
		button:SetNormalFontObject(GameFontNormalSmall)

		button:SetNormalTexture("Interface\\Buttons\\UI-Panel-Button-Up")
		button:SetPushedTexture("Interface\\Buttons\\UI-Panel-Button-Down")
		button:SetHighlightTexture("Interface\\Buttons\\UI-Panel-Button-Highlight")
		button:SetDisabledTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
		button:GetNormalTexture():SetTexCoord(0, 0.625, 0, 0.6875)
		button:GetPushedTexture():SetTexCoord(0, 0.625, 0, 0.6875)
		button:GetHighlightTexture():SetTexCoord(0, 0.625, 0, 0.6875)
		button:GetDisabledTexture():SetTexCoord(0, 0.625, 0, 0.6875)
		button:GetHighlightTexture():SetBlendMode("ADD")

		return button
	end

	local title = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	title:SetPoint("TOPLEFT", 16, -16)
	title:SetText("Project Deranjata LDB Menu")

	local subtitle = frame:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	--~ 	subtitle:SetHeight(32)
	subtitle:SetHeight(35)
	subtitle:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -8)
	subtitle:SetPoint("RIGHT", frame, -32, 0)
	subtitle:SetNonSpaceWrap(true)
	subtitle:SetJustifyH("LEFT")
	subtitle:SetJustifyV("TOP")
	--~ 	subtitle:SetMaxLines(3)
	subtitle:SetText("This panel can be used to configure the Project Deranjata LDB display.")

	local rows, anchor = {}
	local EDGEGAP, ROWHEIGHT, ROWGAP, GAP = 16, 20, 2, 4

	local function OnEnter(self)
		local type = (NinjaPanel.plugins[self.name] and NinjaPanel.plugins[self.name].object.type) or "Unknown"
		GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
		GameTooltip:AddLine(self.name, nil, nil, nil, true)
		GameTooltip:AddLine(type, 1, 1, 1, true)
		GameTooltip:Show()
	end
	local function OnLeave() GameTooltip:Hide() end
	local function OnClick(self)
		local opts = NinjaPanelDB.plugins[self.name]
		local plugin = NinjaPanel.plugins[self.name]
		if opts.disabled then
			opts.disabled = nil
		else
			opts.disabled = true
			NinjaPanel:DetachPlugin(plugin)
		end
		NinjaPanel:ScanForPlugins()
		Refresh()
	end

	-- Create rows for each option
	for i=1,math.floor((305-22)/(ROWHEIGHT + ROWGAP)) do
		local row = CreateFrame("Button", nil, frame)
		if not anchor then row:SetPoint("TOP", subtitle, "BOTTOM", 0, -16)
		else row:SetPoint("TOP", anchor, "BOTTOM", 0, -ROWGAP) end
		row:SetPoint("LEFT", EDGEGAP, 0)
		row:SetPoint("RIGHT", -EDGEGAP*2-8, 0)
		row:SetHeight(ROWHEIGHT)
		anchor = row
		rows[i] = row

		local check = CreateFrame("CheckButton", nil, row)
		check:SetWidth(ROWHEIGHT+4)
		check:SetHeight(ROWHEIGHT+4)
		check:SetPoint("LEFT")
		check:SetNormalTexture("Interface\\Buttons\\UI-CheckBox-Up")
		check:SetPushedTexture("Interface\\Buttons\\UI-CheckBox-Down")
		check:SetHighlightTexture("Interface\\Buttons\\UI-CheckBox-Highlight")
		check:SetDisabledCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check-Disabled")
		check:SetCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check")
		check:SetScript("OnClick", OnClick)
		row.check = check

		local title = row:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
		title:SetPoint("LEFT", check, "RIGHT", 4, 0)
		row.title = title

		local status = row:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
		status:SetPoint("RIGHT", row, "RIGHT", -4, 0)
		status:SetPoint("LEFT", title, "RIGHT")
		status:SetJustifyH("RIGHT")
		row.status = status

		row:SetScript("OnEnter", OnEnter)
		row:SetScript("OnLeave", OnLeave)
	end

	local statusColors = {
		[true] = {0.1, 1.0, 0.1},
		[false] = {1.0, 0.1, 0.1},
	}

	local offset = 0
	Refresh = function()
		if not frame:IsVisible() then return end

		table.sort(NinjaPanel.pluginNames)

		for i,row in ipairs(rows) do
			if (i + offset) <= #NinjaPanel.pluginNames then
				local name = NinjaPanel.pluginNames[i + offset]
				local entry = NinjaPanel.plugins[name]
				local opts = NinjaPanelDB.plugins[name]
				local enabled = not opts.disabled
				local color = statusColors[enabled or false]
                local dataObj = ldb:GetDataObjectByName(name)
                local type = dataObj.type == "launcher" and " |cffffffff(launcher)|r" or ""

				row.check:SetChecked(enabled)
				row.title:SetText(name .. type)
				row.status:SetText(enabled and "Enabled" or "Disabled")
				row.status:SetTextColor(unpack(color))
				row.name = name
				row.check.name = name
				row:Show()
			else
				row:Hide()
			end
		end
	end
	frame:SetScript("OnEvent", Refresh)
	frame:RegisterEvent("ADDON_LOADED")
	frame:SetScript("OnShow", Refresh)
	Refresh()

	local scrollbar = LibStub("NinjaPaneltekKonfig-Scroll").new(frame, nil, #rows/2)
	scrollbar:ClearAllPoints()
	scrollbar:SetPoint("TOP", rows[1], 0, -16)
	scrollbar:SetPoint("BOTTOM", rows[#rows], 0, 16)
	scrollbar:SetPoint("RIGHT", -16, 0)
	scrollbar:SetMinMaxValues(0, math.max(0, #NinjaPanel.pluginNames - #rows))
	scrollbar:SetValue(0)

	local f = scrollbar:GetScript("OnValueChanged")
	scrollbar:SetScript("OnValueChanged", function(self, value, ...)
		offset = math.floor(value)
		Refresh()
		return f(self, value, ...)
	end)

	frame:EnableMouseWheel()
	frame:SetScript("OnMouseWheel", function(self, val) scrollbar:SetValue(scrollbar:GetValue() - val*#rows/2) end)

	local enableall = MakeButton()
	enableall:SetPoint("BOTTOMLEFT", 16, 16)
	enableall:SetText("Enable All")
	enableall:SetScript("OnClick", function(button)
		for idx,name in ipairs(NinjaPanel.pluginNames) do
			NinjaPanelDB.plugins[name].disabled = nil
		end
		Refresh()
	end)

	local disableall = MakeButton()
	disableall:SetPoint("LEFT", enableall, "RIGHT", 4, 0)
	disableall:SetText("Disable All")
	disableall:SetScript("OnClick", function(button)
		for idx,name in ipairs(NinjaPanel.pluginNames) do
			NinjaPanelDB.plugins[name].disabled = true
		end
		Refresh()
	end)

	local reload = MakeButton()
	reload:SetPoint("BOTTOMRIGHT", -16, 16)
	reload:SetText("Refresh")
	reload:SetScript("OnClick", ReloadUI)
end)

if Settings and Settings.RegisterCanvasLayoutCategory then
	local category = Settings.RegisterCanvasLayoutCategory(frame, "Project Deranjata")
	category.ID = "NinjaPanel"
	Settings.RegisterAddOnCategory(category)
elseif InterfaceOptions_AddCategory then
	InterfaceOptions_AddCategory(frame)
end
